/****************************************************************************
*                particle.c
*
*  This module implements the particle primitive.
*
*  from Persistence of Vision(tm) Ray Tracer
*  Copyright 1996,1999 Persistence of Vision Team
*---------------------------------------------------------------------------
Syntax - see particlepatch.html

*/
/*Notes - taken from the report of James Neill

COLOUR EFFECTS
degree of effect f(d) = pd * cropmax(1,	 ps * (d ^ -pf )	)
where
	d = distance from center
	pd = particle density
	ps = " size
	pf = " falloff

ncx =	( 1 - f(d) ) * ocx +
		f(d) * pcx * ocx * (1-ilx) +
		f(di) * pcx * ilx
where terms ending in x are red,green,blue components
	nc = new color of the ray after leaving the particle
	oc = old_colour of the ray before entering the particle
	pc = colour of the particle
	il = illumination component

RAY DIRECTION EFFECTS
ray is bent away/toward particle & | turned left/right (possible to reflect it)

SHADOW EFFECTS
proportional shadows

Trace, Trace_Primary_Ray - refractions inserted here
Do_Finite_Atmosphere - colours inserted here
block_point_light - shadows inserted here - not yet

Note: definition of ParticlePatch in frame.h allows use of these features
*/
#include "frame.h"
#ifdef ParticlePatch
#include <ctype.h>
#include <float.h>
#include "vector.h"
#include "texture.h"
#include "povproto.h"
#include "parse.h"
#include "tokenize.h"
#include "express.h"
#include "povray.h"
#include "particle.h"

/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/
PARTICLE* Default_Particle;/* = Create_Particle();<- done in Parse () */
PARTICLE* Bent_Particle = NULL;

/*kills whitespace in a c FILE   */
void eatwhite(FILE*f){
	/*why the hell is this function not implemented for c FILE[s] */
	/*but implemented for c++ iostream[s] */
	int c;
	while(1){
		c = fgetc(f);
		if((!isspace(c))||c==EOF){
			ungetc(c,f);
			break;
		}
	}
}

int Fix_Format(char*f,char*pf){
	int i,ii;
	for(i=ii=0;f[i]!=0;i++){
		switch(f[i]){
			case'f': case'i': case'a': case's': case'k': case'b': case'd': case'S':
			case'z': case'c': case'C': case'r': case't': case'm': case'l': case'n':
				f[ii]=f[i];ii++;
				break;
			default:
				if(!isspace(f[i])){
					Warning(0,"Illegal format specifier '%c' in 'particles' ... ignoring & continuing (file = \"%s\")",f[i],pf);
				}

		}
	}
	f[ii]=0;
	return ii;
}


PARTICLE* Parse_Particle_File(PARTICLE* TempPart, char*pfilename,char*format){
	PARTICLE* tmp;/*used to stick the next one on the front of the list */
	PARTICLE* New = NULL;/*list of new particles */
	int numin,numtoread,i,getnum;
	DBL dbltmp;
	FILE* pfile;

	/*Locate_File calls Error if pfile is null before returning */
	pfile = Locate_File(pfilename,READ_TXTFILE_STRING,"","",NULL,TRUE);

	/*fix up the format string */
	numtoread = Fix_Format(format,pfilename);

	while(!feof(pfile)){
		/*create a particle on front of list */
		tmp = Create_Particle();
		Copy_Particle(TempPart,tmp);/*copy temp data over */
		tmp->Next = New;
		New = tmp;

		numin = 0;/*haven't read any parameters */

		/*read parameters in the order specified */
		for(i=0;i<numtoread;i++){
			switch(format[i]){
				/*bitflags */
				case'o':
					getnum = fscanf(pfile,"%lf",&dbltmp);if(getnum)numin++;
					SetBitFlag(New->flags,COL_ON,dbltmp);
					break;
				case'r':
					getnum = fscanf(pfile,"%lf",&dbltmp);if(getnum)numin++;
					SetBitFlag(New->flags,REFRAC_ON,dbltmp);
					break;
/*				case's':
					getnum = fscanf(pfile,"%lf",&dbltmp);if(getnum)numin++;
					SetBitFlag(New->flags,SHAD_ON,dbltmp);
					break;
				case'l':
					getnum = fscanf(pfile,"%lf",&dbltmp);if(getnum)numin++;
					SetBitFlag(New->flags,CALC_ILLUM_ON,dbltmp);
					break;
				case'm':
					getnum = fscanf(pfile,"%lf",&dbltmp);if(getnum)numin++;
					SetBitFlag(New->flags,MULT_ILLUM_ON,dbltmp);
					break;*/
				case'I':
					getnum = fscanf(pfile,"%lf",&dbltmp);if(getnum)numin++;
					SetBitFlag(New->flags,INVERT_REFRAC,dbltmp);
					break;

				/*one number */
				case'd':
					getnum = fscanf(pfile,"%lf",&New->density);if(getnum)numin++;
					break;
				case'f':
					getnum = fscanf(pfile,"%lf",&New->falloff);if(getnum)numin++;
					break;
				case'z':
					getnum = fscanf(pfile,"%lf",&New->size);if(getnum)numin++;
					break;
/*				case'n':
					getnum = fscanf(pfile,"%lf",&New->nobenddist);if(getnum)numin++;
					break;*/
				case'k':
					getnum = fscanf(pfile,"%lf",&New->checkrad);if(getnum)numin++;
					break;

				/*2 numbers */
				case'a':
					getnum = fscanf(pfile,"%lf %lf",
						&New->angle[0],
						&New->angle[1]);
					if(getnum==2)numin++;
					break;

				/*3 numbers */
				case'i':
					getnum = fscanf(pfile,"%f %f %f",
						&New->illumcol[0],
						&New->illumcol[1],
						&New->illumcol[2]);
					if(getnum==3)numin++;
					break;
				case'C':
					getnum = fscanf(pfile,"%f %f %f",
						&New->colour[0],
						&New->colour[1],
						&New->colour[2]);
					if(getnum==3)numin++;
					break;
				case'c':
					getnum = fscanf(pfile,"%lf %lf %lf",
						&New->center[0],
						&New->center[1],
						&New->center[2]);
					if(getnum==3)numin++;
					break;
			}
			if((i+1)!=numin)break;/*if we missed a parameter */
		}
		if(numin!=numtoread){
			/*Just to be on the safe side */
			fclose(pfile);
			POV_FREE (format);
			Destroy_Particle(TempPart);
			Error(	"\nError reading parameters from particle file :\"%s\"\n"
									"Format is specified in a string after the file name\n"
									" - see docs for details\n"
									" - check for errant characters (eg a-z,#@%*| etc)"
									"and missing values\n",pfilename);
		}
		eatwhite(pfile);/*get rid of whitespace between particles */
	}

	fclose(pfile);
	return New;
}


PARTICLE* Parse_Particles(void){
	PARTICLE* TempPart;/*store temp particle data */
	PARTICLE* pid;
	PARTICLE* New;/*list of new particles */
	char* pfilename = NULL;
	char* format = NULL;
	TempPart = Create_Particle();
	Copy_Particle(Default_Particle,TempPart);
	New = NULL;

	if(Parse_Begin_If_Present()){/*consume '' */

		opts.particles |= PARTICLES_ON;/*should this be here??? */

		Get_Token();
		if(Token.Token_Id==PARTICLE_ID_TOKEN){
			pid = Duplicate_Particle(Token.Data);
		}
		else{
			pid = NULL;
			UNGET;
		}

		pfilename = Allow_String(FALSE);
		if(pfilename){

			Parse_Comma();/*consume ',' */

			format = Parse_String();

		}

		Parse_Particle_Stuff(pid?pid:TempPart);

		Parse_End();

		/*Now parse the input file */
		if(pfilename){
			if(pid){
				*TempPart = *pid;/*copy data over */
				TempPart->Next = NULL;
				Destroy_Particle(pid);
			}
			New = Parse_Particle_File(TempPart,pfilename,format);
			POV_FREE (pfilename);
			POV_FREE (format);
		}
		else
			New = pid;

	}
	else
		SetBitFlag(opts.particles,PARTICLES_ON,Allow_Float(1));

	Destroy_Particle(TempPart);
	return New;
}

PARTICLE* Parse_Particle(void){
	PARTICLE* New = Create_Particle();
	Copy_Particle(Default_Particle,New);

	if(Parse_Begin_If_Present()){/*consume '{' */

		opts.particles |= PARTICLES_ON;/*should this be here??? */
		Get_Token();
		if(Token.Token_Id==PARTICLE_ID_TOKEN){
			New = Create_Particle();
			Copy_Particle(Token.Data,New);
		}
		else
			UNGET;

		if(Allow_Vector(New->center)){

			Parse_Comma();/*consume ',' */

			New->size = Parse_Float();
/*			Parse_Colour(New->colour);   */

		}

		Parse_Particle_Stuff(New);

		Parse_End();
	}
	else
		SetBitFlag(opts.particles,PARTICLES_ON,Allow_Float(1));

	return New;
}

void Parse_Particle_Stuff(PARTICLE * p){
	VECTOR Local_Vector;
	EXPECT

		CASE(CENTER_TOKEN)
			Parse_Vector(p->center);
		END_CASE

		CASE(TRANSLATE_TOKEN)
			Parse_Vector(Local_Vector);
			VAddEq(p->center,Local_Vector);
		END_CASE

		CASE(SIZE_TOKEN)
			p->size = Parse_Float();
		END_CASE

		CASE(CHECKRAD_TOKEN)
			p->checkrad = Parse_Float();
		END_CASE

		CASE(NOISE_TOKEN)
			p->noise = Parse_Float();
		END_CASE

		CASE(NOISE_SCALE_TOKEN)
			p->noisescale = Parse_Float();
		END_CASE

/*		CASE(LABEL_TOKEN)
			if (p->Label)
				POV_FREE(p->Label);
			Get_Token();
			p->Label = POV_STRDUP(Token.Token_String);
		END_CASE*/

		CASE(DENSITY_TOKEN)
			p->density = Parse_Float();
		END_CASE

		CASE(FALLOFF_TOKEN)
			p->falloff = Parse_Float();
		END_CASE

		CASE(COLOUR_EFFECTS_TOKEN)
			SetBitFlag(p->flags,COL_ON,Allow_Float(1))
		END_CASE

/*		CASE(CALC_ILLUM_TOKEN)
			SetBitFlag(p->flags,CALC_ILLUM_ON,Allow_Float(1))
		END_CASE

		CASE(MULT_ILLUM_TOKEN)
			SetBitFlag(p->flags,MULT_ILLUM_ON,Allow_Float(1))
		END_CASE*/

		CASE_COLOUR
			Parse_Colour( p->colour );
		END_CASE

		CASE(ILLUM_COLOUR_TOKEN)
			Parse_Colour( p->illumcol );
		END_CASE

/*		CASE(SHADOWS_TOKEN)
			SetBitFlag(p->flags,SHAD_ON,Allow_Float(1))
		END_CASE*/

		CASE(REFRACTION_TOKEN)
			SetBitFlag(p->flags,REFRAC_ON,Allow_Float(1))
		END_CASE

		CASE(ANGLE_TOKEN)
			Parse_UV_Vect(p->angle);
		END_CASE

/*		CASE(NOBENDDIST_TOKEN)
			p->nobenddist = Parse_Float();
		END_CASE*/

		CASE(INVERSE_TOKEN)
			SetBitFlag(p->flags,INVERT_REFRAC,Allow_Float(1))
		END_CASE

		OTHERWISE
			UNGET
			EXIT
		END_CASE

	END_EXPECT
}

PARTICLE* Create_Particle (void){
	PARTICLE* New;
	New = (PARTICLE*)POV_MALLOC(sizeof(PARTICLE), "particle");

	Make_Vector(New->center,0,0,0);
	New->size = New->density = New->falloff = New->noisescale = 1.0;
	New->checkrad = New->noise = /*New->nobenddist = */New->angle[1] = 0.0;
	Make_Colour(New->colour,1,1,1);
	Make_Colour(New->illumcol,0,0,0);
	New->angle[0] =  0.1;
	New->flags = 0;
/*	New->Label = NULL; */
	New->Next = NULL;

	return New;
}

void Copy_Particle (PARTICLE* src,PARTICLE* dest){
	memcpy(dest,src,sizeof(PARTICLE));
	dest->Next = NULL;
/*	dest->Label = NULL; */
}

void Destroy_Particle (PARTICLE*p){
	if(p){
		if(p->Next)Destroy_Particle(p->Next);
/*		if(p->Label)POV_FREE(p->Label);   */
		POV_FREE(p);
	}
}

PARTICLE* Duplicate_Particle (PARTICLE*p){
	PARTICLE* New;
	if(!p)return NULL;
	New = Create_Particle();
	*New = *p;
/*	New->Label=NULL; */
	if(p->Next)
		New->Next = Duplicate_Particle(p->Next);
	return New;
}

/*	p1 closest point to p2 on ray, r ray dirm, ro ray-origin, p2 the point we are interested in */
#define VAssign(D,S) {(D)[X] = S[X];(D)[Y] = S[Y];(D)[Z] = S[Z];}
#define VEqEq(A,B) ( ((A)[X] == B[X]) && ((A)[Y] == B[Y]) && ((A)[Z] == B[Z]) )

/*this returns in P the closest point on ray specified by RD (direction) and RO (origin) to point */
#define GetPRC(P,RD,RO,C) \
	{\
		VECTOR V1;DBL L0;DBL L1;\
		VSub(V1,C,RO);\
		VDot(L0,V1,RD);\
		VDot(L1,RD,RD);\
		L0 /= L1;\
		VAddScaled(P,RO,L0,RD);\
	}

int CalcParticleDistances(RAY* Ray){
	PARTICLE*p;/*current particle */
	int i;
	VECTOR tmp;DBL cosA = 0;
	i=0;
	for( p = Frame.Particles; p != NULL ; p = p->Next ){
		i++;
		GetPRC(p->point,Ray->Initial,Ray->Direction,p->center);/*point on ray closest to center*/
		VDist(p->dist,Ray->Initial,p->point);/*don't use this without the next lines - it always returns positive dist we want negative dists as well*/
		/*this bit by Chris Huff (thanks to the newsgroups)*/
		cosA = 0;
		VSub(tmp,p->point,Ray->Initial);
		VNormalizeEq(tmp);
		VDot(cosA, Ray->Direction, tmp);
		if(cosA<0)
			p->dist *= -1;
		/*end by CH*/
		VDist(p->pdist,p->center,p->point);/*dist from center to point*/
	}/*for*/
	return i;
}/*calcdists*/

DBL cropmax(DBL m, DBL v){
	return (v>m?m:v);
}

PARTICLE* GetStart(){
	PARTICLE*p;/* particle*/
	PARTICLE*sp;/*start particle*/
	DBL stval = -DBL_MAX;
	sp = NULL;

	for( p = Frame.Particles; p!=NULL ;p = p->Next){
		if (p->dist < 0 && p->dist > stval){
			sp = p;stval = p->dist;
		}
	}

	return sp;
}

PARTICLE* GetEnd(DBL min){
	PARTICLE*p;/* particle*/
	PARTICLE*ep;/*start particle*/
	DBL val = DBL_MAX;
	ep = NULL;

	for( p = Frame.Particles; p!=NULL ;p = p->Next){
		if (p->dist > min && p->dist < val){
			ep = p;val = p->dist;
		}
	}

	return ep;
}

PARTICLE* GetClosestParticle(RAY* r, DBL intersecdepth){
	PARTICLE* closest_particle;/*pointer to particle we want*/

	PARTICLE*p;/* particle*/
	DBL mdist;/*min/max dist*/

	mdist = 0 ;
	closest_particle = NULL;

	for( p = Frame.Particles; p!=NULL ;p = p->Next){
		if(		( opts.particles&PARTICLES_ON ) &&				/*particles are on*/
				( ! (p->flags&GET_CHECKED) ) &&					/* get particles where the checked flag is off*/
				( p->flags&COL_ON) &&							/*particles using colour effects*/
				( p->checkrad ? p->pdist < p->checkrad : 1) &&	/*inside checkrad*/
				( mdist < p->dist ) &&							/*get the max*/
				( p->dist > 0 ) &&								/*between ray origin*/
				( p->dist < intersecdepth )						/* & intersection depth*/
			){
			mdist = p->dist;
			closest_particle = p;
		}/*if*/
	}/*for*/

	if(!closest_particle && opts.particles&INCEND_ON){
		VECTOR tmp;
		closest_particle=GetEnd(intersecdepth);
		if(closest_particle){
			VAssign(tmp, r->Direction);
			VNormalizeEq(tmp);
			VScaleEq(tmp,intersecdepth);
			VAdd(closest_particle->point,r->Initial,tmp);
			closest_particle->dist = intersecdepth;
			VDist(closest_particle->pdist,closest_particle->center,closest_particle->point);/*dist from center to point*/
		}
	}
	if(!closest_particle && opts.particles&INCSTART_ON){
		closest_particle=GetStart();
		if(closest_particle){
			VAssign(closest_particle->point, r->Initial);closest_particle->dist = 0;
			VDist(closest_particle->pdist,closest_particle->center,closest_particle->point);/*dist from center to point*/
		}
	}

	if(closest_particle)closest_particle->flags |= GET_CHECKED;
	return closest_particle;
}/*getclosestparticle*/

/*inline because only used twice - shadow func, color func*/
#ifndef __inline
#define __inline
#endif

__inline DBL func(PARTICLE*p){
	VECTOR tmp;
	VScale(tmp,p->point,p->noisescale);
	return p->density * cropmax( 1, p->size / pow( p->pdist , p->falloff ) ) + ( p->noise ? p->noise * Noise(tmp) : 0 );
}

void do_particle_colours(RAY *Ray, INTERSECTION *Intersection, COLOUR Colour){
	PARTICLE*p;/*current particle*/
	DBL fd;/*3d function*/
	int x, nump,i;/*used to assign colours*/

	/*we don't actually want to change the intersection
	  but we may need to alter the intersection during processing
	  .'. make a copy of the intersection  */
	INTERSECTION Intersec;
	Intersec = *Intersection;

	nump = CalcParticleDistances(Ray);

	for( p = Frame.Particles; p!=NULL ;p = p->Next)
		p->flags&=~GET_CHECKED;/*turn off checkeds*/

	for(i = 0;i < nump; i++){/*	while(1){//	while(p = GetClosestParticle(Ray,Intersection->Depth)){/*
		/*can anyone see why using a while loop here would*/
		/*(taking into account the next two lines and the functions GetClosestParticle & CalcParticleDistances)*/
		/*cause an infinite loop to occur*/
		p = GetClosestParticle(Ray,Intersection->Depth);
		if(p==NULL)break;/*quit if we get a bad particle*/

		/*alter this to change the pattern - when support for isosurface functions is added this may have to be translated by p->center*/
		fd = func(p);

		/*calc colour effects - I recommend that more methods of calculating this and fd are created*/
		for(x=0;x<3;x++){
			Colour[x] =	(COLC)( 1.0 - fd ) * Colour[x] + /*old colour of ray*/
						(COLC)fd * p->colour[x] * Colour[x] * ((COLC)1.0 - p->illumcol[x]) +/*mixture between old and illumination colours*/
						(COLC)fd * p->colour[x] * p->illumcol[x];/*illumination of ray*/
		}/*color application loop*/
	}/*loop*/
}/*do_p_colors*/

PARTICLE* GetClosestBend(RAY* r, DBL intersecdepth){
	PARTICLE* closest_particle;/*pointer to particle we want*/

	PARTICLE*p;/* particle*/
	DBL mdist;/*min/max dist*/

	mdist = intersecdepth ;
	closest_particle = NULL;

	for( p = Frame.Particles;p!=NULL;p = p->Next){
		if(		( opts.particles&PARTICLES_ON ) &&				/*particles are on*/
				( p->checkrad ? p->pdist < p->checkrad : 1) &&	/*inside checkrad*/
				( mdist > p->dist ) &&							/*get the min*/
				( p->dist > 0 ) &&								/*between ray origin*/
				( p->dist < intersecdepth ) &&					/* & intersection depth*/
				( p->flags&REFRAC_ON) &&						/*particles using refraction*/
				( p != Bent_Particle )						/*ignore particles that have bent the ray*/
			){

			mdist = p->dist;
			closest_particle = p;
		}/*if*/

	}/*for*/

	return closest_particle;
}/*getclosestbend*/

char BendRay(RAY *Ray, INTERSECTION* Intersection){
	PARTICLE* p;
	DBL ra[2] = {0,0};/*refract amount*/
	VECTOR ptor[2];
	VECTOR tmp;
	char b = 0;
	CalcParticleDistances(Ray);
	p = GetClosestBend(Ray,Intersection->Depth);
	if(p){
		VAssign(Ray->Initial, p->point);
		if(p->angle[0]){
			if(p->noise)VScale(tmp,p->point,p->noisescale);
			ra[0] = ( p->pdist ? p->angle[0] * p->size / p->pdist : 0 ) + ( p->noise ? p->noise * Noise(tmp) : 0 );/* - p->angle[0] * p->size / p->nobenddist * exp( p->pdist - p->nobenddist )  );*/
		}
		if(ra[0] || p->angle[1]){
			if(p->flags&INVERT_REFRAC){ VSub(ptor[0],p->center,p->point);}
			else{ VSub(ptor[0],p->point,p->center);}
			VNormalizeEq(ptor[0]);
			if(ra[0])VAddScaledEq(Ray->Direction, ra[0], ptor[0]);
		}
		if(p->angle[1]){
			ra[1] = ( p->pdist ? p->angle[1] * p->size / p->pdist: 0 ) + ( p->noise ? p->noise * Noise(tmp) : 0 );/* - p->angle[1] * p->size / p->nobenddist * exp( p->pdist - p->nobenddist )  );*/
			if(ra[1]){
				VCross(ptor[1],Ray->Direction,ptor[0]);
				VNormalizeEq(ptor[1]);
				VAddScaledEq(Ray->Direction, ra[1], ptor[1]);
			}
		}
		if(ra[0] || ra[1]){
			VNormalizeEq(Ray->Direction);
			Bent_Particle = p;/*fix so that the ray doesn't get re-bent by the same particle, causing stack overflow*/
		}
		return TRUE;/*did bend ray*/
	}
	return FALSE;/*didn't bend ray*/
}

/*DBL Particle_Shadow_Influence(RAY* Ray, INTERSECTION* Intersection, VECTOR pt1, VECTOR pt2){
	PARTICLE* p;
	DBL totalfd = 0, d = 0;
	CalcParticleDistances(Ray);
	for( p = Frame.Particles;p!=NULL;p = p->Next){
		if(p->dist > 0.0 && p->dist < Intersection->Depth && p->dist < p->checkrad){
			totalfd += func(p);
		}
	}
	return cropmax(1,totalfd);
}*/

#endif /*ifdef ParticlePatch*/
