/****************************************************************************
*                   partsys.h
*
*  This module contains all defines, typedefs, and prototypes for partsys.c.
*   Author: Chris Huff
*
*****************************************************************************/


#ifndef PARTSYS_H
#define PARTSYS_H


/*****************************************************************************
* Global preprocessor defines
******************************************************************************/

#define PARTICLE_SYSTEM_OBJECT (BASIC_OBJECT)

/*****************************************************************************
* Global typedefs
******************************************************************************/

typedef enum
{
    ATMOSPHERE_PARTICLE = 0,
    FLUID_PARTICLE
} PARTICLE_TYPE;

typedef enum
{
    GRAVITY_SOURCE = 0,
    LOCAL_WIND,
    TWIRL
} SPACE_WARP_TYPE;

typedef enum
{
    POINT_EMITTER = 0,
    OBJECT_EMITTER
} EMITTER_TYPE;

typedef enum
{
    NO_IPF = 0,
    CUSTOM_IPF,
    IPF_A
} IPF_PROFILE;

typedef struct Particle_System_Object_Struct PART_SYS_OBJ;
/*typedef struct Particle_System_Struct PART_SYS;*/
typedef struct Emitter_Struct PART_EMITTER;
typedef struct Part_Struct PART;
typedef struct Space_Warp_Struct SPACE_WARP;



struct Particle_System_Object_Struct
{
  OBJECT_FIELDS

  /*Blob stuff*/
  TRANSFORM *Trans;
  BLOB_DATA *Data; /* Pointer to blob data  */
  TEXTURE **Element_Texture;

  /*    OBJECT * obj;*/
  PART_SYS * sys;
};

struct Particle_System_Struct
{
    unsigned long int references;

    unsigned int max_bounces;

    DBL drag;
    DBL tScale;
    PIGMENT *temperature_pigment;
    OBJECT *env;
    VECTOR grav;
    VECTOR wind;
    VECTOR jitter;

    int seed;
    DBL time;

    unsigned long int num_of_particles;

    unsigned long int particles_created;
    unsigned long int particles_killed;
    unsigned long int particle_drag_interactions;
    unsigned long int particle_bond_interactions;
    unsigned long int particle_collisions;
    unsigned long int num_of_emitters;
    unsigned long int num_of_space_warps;
    DBL max_velocity;
    PART_EMITTER *emitter_list;
    SPACE_WARP *space_warp_list;
    PART *particle_list;
};

#define PART_FIELDS           \
    PARTICLE_TYPE type;       \
/*    unsigned int max_bounces;*/ \
    DBL radius;               \
    DBL strength;               \
    DBL elasticity;           \
    SNGL age;                 \
    DBL lifespan;             \
    DBL mass;                 \
    SNGL temperature;         \
    VECTOR pos;               \
    VECTOR vel;               \
    DBL mass_jitter;          \
    DBL lifespan_jitter;

struct Emitter_Struct
{
    PART_FIELDS

    IPF_PROFILE ipf_type;
    DBL ipf_strength;
    DBL ipf_exp;
    DBL ipf_dist;
    DBL ipf_drag;
    DBL ipf_search_rad;
    SPLINE2 *ipf_spline;

    DBL particle_stretch;
    DBL particle_stretch_min;
    DBL emission_rate;
    DBL min_angle;
    DBL max_angle;
    DBL start_time;
    DBL stop_time;
    EMITTER_TYPE emitter_type;
    OBJECT *obj;
    int max_tests;
    TRANSFORM *trans;
    TEXTURE *texture;

    DBL last_emitted;
    PART_EMITTER *next;
};

struct Space_Warp_Struct
{
    SPACE_WARP_TYPE type;
    VECTOR pos;
    VECTOR vect_dat_a;
    DBL radius;
    DBL strength;
    DBL falloff;
    TPATTERN *warp;
    TRANSFORM *trans;

    SPACE_WARP *next;
};

struct Part_Struct
{/*Only age, temperature, position, and velocity are stored in the particle.
The rest of the values don't change, so they are stored in the emitter which
created the particle.*/
/*    unsigned int bounces;*/
    SNGL age;
    SNGL temperature;
    VECTOR oldPos;
    VECTOR oldVel;
    VECTOR pos;
    VECTOR vel;
    PART_EMITTER *parent_emitter;
    PART *prev;
    PART *next;
};

/*****************************************************************************
* Global variables
******************************************************************************/



/*****************************************************************************
* Global functions
******************************************************************************/


PART_SYS_OBJ * Create_Part_Sys_Obj(void);
PART_SYS_OBJ * Copy_Part_Sys_Obj(OBJECT * Object);
void Destroy_Part_Sys_Obj(OBJECT * Object);

PART_SYS * Create_Particle_System();
PART_SYS * Copy_Particle_System(PART_SYS * src);
void Destroy_Particle_System(PART_SYS * sys);
void Destroy_Particle_Emitter_List(PART_SYS * sys);
void Destroy_Space_Warp_List(PART_SYS * sys);
PART_EMITTER *  Create_Particle_Emitter();
void Add_Particle_Emitter(PART_SYS * sys, PART_EMITTER * em);
SPACE_WARP * Create_Space_Warp();
void Add_Space_Warp(PART_SYS * sys, SPACE_WARP * wp);

void Run_Sys(PART_SYS * sys, unsigned long int cycles, DBL sim_length);

PART * Get_Particle(PART_SYS * sys, unsigned long int idx);

void Output_Particle_System_Statistics(PART_SYS * sys);

DBL Calc_Field_Strength(PART_SYS * sys, VECTOR EPoint);

#endif
