/****************************************************************************

FILE	syms.c

PURPOSE 
    This file contains routines for handling symbols.

*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "shade.h"
#include "y.tab.h"
#include "opcodes.h"
#include "shdrglob.h"
#include "shmem.h"

#define HSIZE   101

int shaderstart = 0;

struct _sym *symtab[HSIZE];
static struct func_tree *first_function=0;


static struct _sym * dfltsym(char *s, int ty, int offset);

struct _cnst {
    int type;
    int offset;
    int reloc_offset;
    union {
        double f;
        char *s;
    } v;
    struct _cnst *next;
};

#define  MAX_CONTEXT 50

struct context{
    int type; /* func or shader */
    char * name;
    int block;  /* starts from 1! */
    int dot;
    int block_start[MAX_CONTEXT];
};


struct loop_struct
{
    int break_offs;
    int continue_offs;
    struct ilist *breaks;
    struct ilist *continues;
};

static struct loop_struct loop_arr[MAX_CONTEXT];

static struct context  context_arr[MAX_CONTEXT];
static int context_index=-1;


struct _cnst *cnsts, **nxtcnst;
int nstrings;

#define ACT_SYM_STACK_SIZE 1000
struct _sym * active_sym_stack[ACT_SYM_STACK_SIZE];
int active_count =-1;


struct builtin {
    char *name;
    int type;
    int offset;
};


struct builtin bivars[] = {
    {"Cs", tCOLOR,  Cs_DA_OFFS},
    {"Os", tCOLOR,  Os_DA_OFFS},
    {"P", tPOINT,   P_DA_OFFS},
    {"N", tNORMAL,  N_DA_OFFS},
    {"Ng", tNORMAL, Ng_DA_OFFS},
    {"I", tVECTOR,  I_DA_OFFS},
    {"Ci", tCOLOR,  Ci_DA_OFFS},
    {"Oi", tCOLOR,  Oi_DA_OFFS},
    {"s", tFLOAT,   s_DA_OFFS},
    {"t", tFLOAT,   t_DA_OFFS},
    {"u", tFLOAT,   u_DA_OFFS},
    {"v", tFLOAT,   v_DA_OFFS},
    {"du", tFLOAT,  du_DA_OFFS},
    {"dv", tFLOAT,  dv_DA_OFFS},
    {"time", tFLOAT,time_DA_OFFS},
    {"E", tPOINT,   E_DA_OFFS},
    {"Cl", tCOLOR, Cl_DA_OFFS},
    {"L", tVECTOR, L_DA_OFFS},
    {"As", tCOLOR, As_DA_OFFS},
    {"iso_val", tFLOAT, iso_val_DA_OFFS},
};

#define NBIVARS     (sizeof(bivars)/sizeof(bivars[0]))


void add_active_symbol(struct _sym * sym)
{
    active_count++;
    if (active_count >= ACT_SYM_STACK_SIZE){
        yyerror("Active symbol stack overflow.\n");
    }
    active_sym_stack[active_count]=sym;
}

void remove_active_symbols(int index)
{
    active_count=index;
    if (active_count < 0){
        yyerror("Error in active symbol removing: invalid index.\n");
    }
}


char * get_context_name(void)
{
    return context_arr[context_index].name;
}


void set_context(char * context_name)
{

    struct context * ctx;

    context_index++;
    if (context_index == MAX_CONTEXT){
        char error_buf[100];
        sprintf(error_buf, "Maximum context nesting %d has been reached.\n", MAX_CONTEXT);
        yyerror(error_buf);
    }
    ctx = &(context_arr[context_index]);    
    ctx->name=context_name;
    ctx->block=1;
    ctx->type = 0;
    ctx->dot = 0;
    ctx->block_start[(ctx->block)-1]=active_count;
}


void remove_context(void)
{
    struct context * ctx = &(context_arr[context_index]);    
    remove_active_symbols(ctx->block_start[0]);
        
    if (context_index>=0)
        --context_index;
    else
        yyerror("Error in context removing function.\n");
}


void enter_block(void)
{
    struct context * ctx = &(context_arr[context_index]);    
    (ctx->block)++;
    if (ctx->block > MAX_CONTEXT){
        char error_buf[100];
        sprintf(error_buf, "Maximum block nesting %d has been reached.\n", MAX_CONTEXT);
        yyerror(error_buf);
    }
    ctx->block_start[(ctx->block)-1]=active_count;

}

void leave_block(void)
{
    struct context * ctx = &(context_arr[context_index]);    
            
    remove_active_symbols(ctx->block_start[(ctx->block)-1]);
    (ctx->block)--;
    if (ctx->block <=0){
        yyerror("Error in block ending function.\n");
    }    
}

int get_active_context_start(void)
{
    struct context * ctx = &(context_arr[context_index]);    
    return ctx->block_start[0];
}


int get_active_block_start(void)
{
    struct context * ctx = &(context_arr[context_index]);    
    return ctx->block_start[(ctx->block)-1];
}

int get_block(void)
{
    return context_arr[context_index].block;
}

static int dot=0;

int get_dot(void)
{
    return dot;
}

void set_dot(int new_dot)
{
    dot = new_dot;
}

int hash(char *s)
{
    unsigned int r = 0;
    while (*s != '\0')
        r = (r + *s++) % HSIZE;
    return r;
}

static int func_index = 0;
struct func_tree* install_function(char *name, int type, int offset, int start, int end)
{
    struct func_tree *new_function = sh_malloc(sizeof(struct func_tree));
    new_function->name=name;
    new_function->type=type;
    new_function->code_start=start;
    new_function->code_end=end;
    new_function->next=first_function;
    new_function->off=offset;
    new_function->first_param=NULL;
    new_function->func_index = func_index++;
    new_function->ret_patches = NULL;
    new_function->func_code = NULL;
    first_function=new_function;
    return new_function;
}

struct func_tree *get_function(char *name){
    struct func_tree *ret_val=0;
    for(ret_val=first_function; ret_val; ret_val=ret_val->next){
            if (strcmp(name, ret_val->name)==0){
                break;
            }
        }
        return ret_val;
}

struct func_tree *get_function_by_offset(int offset)
{
    struct func_tree *ret_val=0;
    for(ret_val=first_function; ret_val; ret_val=ret_val->next){
        if (ret_val->code_start == offset){
            break;
        }
    }
    return ret_val;
}

void add_param(struct func_tree * function, struct _sym * param)
{

    struct func_param * new_param = sh_malloc(sizeof(struct func_param));
    struct func_param * last_param = function -> first_param;
    new_param -> param = param;
    new_param -> next_param = last_param;
    function -> first_param = new_param;
}

struct tree * fix_ret_expr(struct tree * ret_val){

    if (ret_val!= NULL){
        struct tree * new_expr = node2(ret_val->op, ret_val->ty, NULL, NULL);
        
        if (ret_val->off != -1){
            struct _sym * var = getsym(ret_val->off);
            if (var!=NULL && var->repl_offset>=0){
                new_expr->off = var->repl_offset;
            }
            else{
                new_expr->off = ret_val->off;
            }
        }    
        new_expr->kids[0] = fix_ret_expr(ret_val->kids[0]);
        new_expr->kids[1] = fix_ret_expr(ret_val->kids[1]);
        new_expr->kids[2] = fix_ret_expr(ret_val->kids[2]);
        new_expr->kids[3] = fix_ret_expr(ret_val->kids[3]);
        return new_expr;
    }
    else{
        return NULL;
    }
}

/*
    This function performs given function inlining.
    It is done by copying function code and fixing jump and variable offsets.
*/

void insert_function(char *name)
{
    struct func_tree *function = get_function(name);
    int counter = function -> code_start;
    int offset = get_dot() - counter - 1;
    int current_op = code[counter];
    int recurse_counter = 1;
    while (current_op != oFSTART){
        int copy_count = get_op(current_op)->size;
        while(copy_count--){
            current_op = code[++counter];
        } 
    }            
    counter+= get_op(oFSTART)->size;
    current_op = code[counter];
        
    while (current_op != oRET){
        int offset_val = is_offset_op(current_op);
        if (is_branch_op(current_op)){
            /* NB! assumption, that each branch operator is 3 bytes! */
            int cur_offset = code[counter+2] |(code[counter+1]<<8);
            cur_offset += offset;
            emit2(current_op, cur_offset);
            counter += get_op(current_op)->size;
        }
        else if (offset_val){
            int cur_offset = code[counter+offset_val+1] |(code[counter+offset_val]<<8);
            int new_offset=cur_offset;
            struct func_param *param = function->first_param;
            int op_length = get_op(current_op)->size;
            while (param != NULL){
                if (param->param->offset == cur_offset && 
                    param->param->repl_offset >= 0){
                    new_offset = param->param->repl_offset;
                    break;
                }
                param = param->next_param;
            }
            if (param != NULL && param->array_index != NULL){
                emit(param->array_index->kids[0]);
                if (istriple(param->array_index->ty)){
                    emit1(oARR3);
                }
                else{
                    emit1(oARR);
                }
                emit2(current_op, new_offset);
            }
            else{
                emit2(current_op, new_offset);
            }
            if (op_length > 3){
                /* copy remaining of operator arguments: 3 was copied by
                    emit2 above 
                */
                int to_copy;
                for(to_copy=0; to_copy < op_length - 3; to_copy++){
                    emit1(code[counter+offset_val+ 2 + to_copy]);
                }
            }
            counter += op_length;
        }
        else if (current_op == oFSTART){
            int cur_dot = counter;
            recurse_counter=1;
            while (recurse_counter > 0){
                int copy_count = get_op(current_op)->size;
                counter += copy_count;
                current_op = code[counter];
                if (current_op == oFSTART){
                    recurse_counter ++;
                }
                else if (current_op == oRET){
                    recurse_counter--;
                }                    
            }
            counter += get_op(current_op)->size;
            offset -= counter - cur_dot;
        }
        else{
            int copy_count = get_op(current_op)->size;
            while(copy_count--){
                emit1(current_op);
                current_op = code[++counter];
            } 
        }
        current_op = code[counter];
    }
}

void add_ret_stmt(struct func_tree * func, struct ilist * stmt)
{
    stmt->next_ilist = func->ret_patches;
    func->ret_patches = stmt;
}

static int loop_count = -1;

/*
    Startup function for loops, sets up structures to handle break and
    continue statements with arbitrary level values.
*/
void start_loop(void)
{
    loop_count ++;
    if (loop_count ==MAX_CONTEXT){
        char error_buf[100];
        sprintf(error_buf, "Maximum loop nesting %d has been  reached.\n", MAX_CONTEXT);
        yyerror(error_buf);
    }
    else{
        loop_arr[loop_count].breaks=NULL;
        loop_arr[loop_count].continues=NULL;
        loop_arr[loop_count].break_offs=0;
        loop_arr[loop_count].continue_offs=0;
        
    }
}

void end_loop(void)
{

    if (loop_count >= 0){
        struct loop_struct * cur_str = &loop_arr[loop_count];
        backpatch(cur_str->breaks, cur_str->break_offs);
        backpatch(cur_str->continues, cur_str->continue_offs);
        loop_count--;
    }
    else{
        yyerror("Error in loop handling ending\n");
    }
}

void set_break(void)
{
    if (loop_count >= 0){
        struct loop_struct * cur_str = &loop_arr[loop_count];
        cur_str->break_offs = get_dot();
    }
}

void set_continue(void)
{
    if (loop_count >= 0){
        struct loop_struct * cur_str = &loop_arr[loop_count];
        cur_str->continue_offs = get_dot();
    }
}

int reg_break(struct ilist * patch, int break_level)
{
    int offs = loop_count-break_level;
    /* convert from 0 -base loop count */
    offs++;
    if (offs<0){
        /* user requires to break out from more loops than exists ! */
        return 0;
    }
    else{
        patch->next_ilist = loop_arr[offs].breaks;
        loop_arr[offs].breaks = patch;
    }
    return 1;
}

int reg_continue(struct ilist * patch, int continue_level)
{
    int offs = loop_count-continue_level;
    /* convert from 0 -base loop count */
    offs++;
    if (offs<0){
        /* break more loops that exists ! */
        return 0;
    }
    else{
        patch->next_ilist = loop_arr[offs].continues;
        loop_arr[offs].continues = patch;
    }
    return 1;
}


struct _sym * install(char *s, int ty, int off, int flags, int  dim)
{

    int slot = hash(s);
    struct _sym *p = NULL;
    int act_index;
    
    if (p == NULL) {
        p = (struct _sym *)sh_malloc(sizeof(struct _sym));
        p->name = s;    /* already saved in gram.y */
        p->type = ty;
        p->context = get_context_name();
        p->offset = off;
        p->flags = flags;
        p->dim = dim;
        p->repl_offset=-1;
        p->reloc_offset=-1;
        if (flags & (S_BUILTIN | S_PARAM)){
            /* 0 indicates builtin or shader parameter */
            p-> block_count = 0;
        }
        else{
            p-> block_count = get_block();
        }
        p->next = symtab[slot];
        symtab[slot] = p;
        add_active_symbol(p);
        if (ty == tSTRING){
            /* don't relocate strings */
            p->reloc_offset = off;
            p->flags |= S_USED;
        }
    }
    return p;
}


int install_fconst(double f)
{
    extern int cur_offset;
    struct _cnst *c;

    for(c = cnsts; c != NULL; c = c->next)
        if (c->type == tFLOAT && c->v.f == f)
            break;
    if (c != NULL){
        return c->offset;
    }

    c = (struct _cnst *)sh_malloc(sizeof(struct _cnst));
    c->type = tFLOAT;
    c->v.f = f;
    c->offset = cur_offset;
    c->reloc_offset = -1;
    c->next = cnsts;
    cnsts = c;

    cur_offset++;

    return c->offset;
}

int is_constant(int offset)
{
    struct _cnst *c;

    for(c = cnsts; c != NULL; c = c->next)
        if ( c->offset  == offset )
            return 1;
    return 0;                        
}

double getfloat(int offset)
{
    struct _cnst *c;

    for(c = cnsts; c != NULL; c = c->next)
        if ( c->offset  == offset && c->type == tFLOAT)
            return c->v.f;
    return 0.0;

}

char * getstring(int offset)
{
    struct _cnst *c;

    for(c = cnsts; c != NULL; c = c->next)
        if ( c->offset  == offset && c->type == tSCONST)
            return c->v.s;
    return "";
}

struct _sym *getsym(int offset)
{
    struct _sym *p;
    int slot;
    for ( slot = 0; slot < HSIZE; ++slot){
        for(p = symtab[slot]; p != NULL; p = p->next){
            if (p->type == tSTRING)
                continue;
            if (p->offset == offset){
                return p;
            }
            else if (p->dim >1 && p->offset < offset){
                /* check for array element */
                int array_dim = p->dim;
                if (istriple(p->type)){
                    array_dim *= 3;
                }
                if (offset < p->offset + array_dim)
                    return p;
            }
        }
    }
    return NULL;
}


struct _sym *get_string_sym(int offset)
{
    struct _sym *p;
    int slot;
    for ( slot = 0; slot < HSIZE; ++slot){
        for(p = symtab[slot]; p != NULL; p = p->next){
            if (p->type != tSTRING)
                continue;
            if (p->offset == offset){
                return p;
            }
        }
    }
    return NULL;
}


int install_sconst(char *s)
{
    struct _cnst *c;

    for(c = cnsts; c != NULL; c = c->next)
        if (c->type == tSCONST && strcmp(c->v.s, s) == 0)
            break;
    if (c != NULL)
        return c->offset;

    c = (struct _cnst *)sh_malloc(sizeof(struct _cnst));
    c->type = tSCONST;
    c->v.s = s;
    c->offset = nstrings++;
    c->reloc_offset = c->offset;
    c->next = cnsts;
    cnsts = c;

    return c->offset;
}


int get_string_offset()
{
    return nstrings++;
}


struct _sym * lookup(char *s)
{
    int end = get_active_block_start();
    struct _sym *p = NULL;
    int act_index;
    for (act_index = active_count; act_index >end; act_index--){
        struct _sym *pp = active_sym_stack[act_index];
        if (strcmp(s, pp->name) ==0){
            p = pp;
            break;
        }
    }        
    return p;
}

struct _sym * lookup_cur_context(char *s)
{
    int end = get_active_context_start();
    struct _sym *p = NULL;
    int act_index;
    for (act_index = active_count; act_index >end; act_index--){
        struct _sym *pp = active_sym_stack[act_index];
        if (strcmp(s, pp->name) ==0){
            p = pp;
            break;
        }
    }        
    return p;
}


struct _sym * lookup_prev_decl(char *s)
{
    struct _sym *p = NULL;
    int act_index;
    for (act_index = active_count; act_index >=0; act_index--){
        struct _sym *pp = active_sym_stack[act_index];
        if (strcmp(s, pp->name) ==0){
            p = pp;
            break;
        }
    }        
    return p;
}

static int illuminance_loop =0;
void start_illuminance(void)
{
    enter_block();
    illuminance_loop = 1;
}

void end_illuminance(void)
{
    leave_block();
    illuminance_loop = 0;
}

int in_illuminance(void)
{
    return illuminance_loop;
}

struct _sym * dfltsym(char *s, int ty, int offset)
{
    extern int cur_offset;
    struct _sym * new_sym;
    
    assert(cur_offset == offset);
    
    new_sym = install(s, ty, offset, S_BUILTIN, 1);
    relocate_sym(cur_offset);
    
    if ( istriple(ty))
        cur_offset += 3;
    else
        cur_offset++;
    return new_sym;
}

static int relocation_offset = 0;


int relocate_sym(int offset)
{
    struct _sym * cur_sym = getsym(offset);
    if (cur_sym){
        if ( (cur_sym->flags & S_USED) == 0){
            int sym_size;
            cur_sym-> flags |= S_USED;
            if ( istriple(cur_sym->type))
                sym_size = 3;
            else if (cur_sym->type == tFLOAT)
                sym_size = 1;
            cur_sym->reloc_offset = relocation_offset;
            relocation_offset += sym_size * cur_sym->dim;
        }
        return cur_sym->reloc_offset + offset - cur_sym->offset;
    }
    else if (is_constant(offset)){
        struct _cnst *c;
        for(c = cnsts; c != NULL; c = c->next){
            if ( c->offset  == offset){
                if( c->type != tSCONST){
                    if ( c->reloc_offset == -1){
                        c->reloc_offset = relocation_offset;
                        relocation_offset += 1;
                    }
                    return c->reloc_offset;
                }
                else{
                    return c->offset;
                }
            }
        }
    }
    else{
        fprintf(stderr, "invalid relocation offset  %d found\n", offset);
        return 0;
    }
}

void end_sym_relocation(void)
{
    struct _sym *p;
    struct _cnst *c;
    int slot;
    
    for ( slot = 0; slot < HSIZE; ++slot){
        for(p = symtab[slot]; p != NULL; p = p->next){
            p->offset = p->reloc_offset;
        }
    }
    for(c = cnsts; c != NULL; c = c->next){
        c->offset = c->reloc_offset;
    }
}


void sym_initialize(void)
{
    int i;
    struct _sym *illum_sym;
    set_context("GLOBAL");
    for(i = 0; i < NBIVARS; i++)
        dfltsym(bivars[i].name, bivars[i].type, bivars[i].offset);

    illum_sym = lookup_prev_decl("Cl");
    illum_sym->flags = illum_sym->flags | F_ILLUMIN;
    illum_sym = lookup_prev_decl("L");
    illum_sym->flags = illum_sym->flags | F_ILLUMIN;
}


int uv_used(void)
{
    struct _sym *uvval = lookup_prev_decl("u");
    if (uvval->flags & S_REFERENCED)
        return 1;
    uvval = lookup_prev_decl("v");
    if (uvval->flags & S_REFERENCED)
        return 1;
    uvval = lookup_prev_decl("s");
    if (uvval->flags & S_REFERENCED)
        return 1;
    uvval = lookup_prev_decl("t");
    if (uvval->flags & S_REFERENCED)
        return 1;
    return 0;
}

char * type_to_string(int type){
    return type == tCOLOR ? "color" :
        type == tPOINT ? "point":
            type == tVECTOR ? "vector":
                type == tNORMAL ? "normal" : 
                    type == tFLOAT ? "float" : "string";

}

void dumpsyms(void)
{
    extern int cur_offset;
    struct _sym *p;
    struct _cnst *c;
    struct func_tree *f;
    int n;

    fprintf(output_file,"; version %d.%d\n", SHADER_VERSION_MAJOR, SHADER_VERSION_MINOR);
    fprintf(output_file,"; shadername %s %d\n", shadername, shaderstart);
    fprintf(output_file,"; shadertype %d\n", shadertype);
    fprintf(output_file,"; requires %d\n", relocation_offset);
    fprintf(output_file,"; codesize %d\n", calc_codesize(&code[0], get_dot()));
    fprintf(output_file,"; numstrings %d\n", nstrings);
    if (uv_used())
        fprintf(output_file, "; uvused\n");
    for(n = 0; n < HSIZE; n++){
        for(p = symtab[n]; p != NULL; p = p->next) {
            if ( (p->flags & S_USED) != 0 && (p->flags & S_BUILTIN) == 0){
                fprintf(output_file,"; sym %s %s %d %d",
                    p->name, type_to_string( p->type),
                    p->offset, p->dim);
                if (p->block_count >= 0)
                    fprintf(output_file," %d", p->block_count );
                fputc('\n', output_file);
            }
        }
    }
    for(c = cnsts; c != NULL; c = c->next){
        if (c->type == tFLOAT){
            if (c->offset != -1){
                fprintf(output_file,"; const float <%g> %d\n", c->v.f, c->offset);
            }
        }
        else if (c->type == tSCONST){
            fprintf(output_file,"; const string %s %d\n",  c->v.s, c->offset);
        }
    }
    if (output_functions){
        for (f = first_function; f != NULL; f = f->next){
            int counter = 1;
            struct func_param * param = f -> first_param;
            fprintf(output_file,"; func %s, start address %d, returns %s @ %d.\n",
                f -> name, f->code_start, type_to_string(f -> type), f->off);
            while (param != NULL){
                fprintf(output_file,";\tParam #%d: offset %d, type: %s\n", 
                    counter, param->param->offset, type_to_string(param->param->type) );
                param = param -> next_param;
                ++ counter;
            }
        }
    }
}





