/****************************************************************************

FILE	shdrread.c

PURPOSE 
    	This file implements shader reading from file.
	
*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/
#include "frame.h"
#ifdef ShaderPatch

#include <stdio.h>
#include "shdrvars.h"
#include "mem.h"
#include "povproto.h"
#include "povray.h"

shader_t *shaders[HSIZE];
static int nloaded = 0;

/*
 * This hashing stuff is primarily for quickly looking up
 * symbols when instantiating shaders.  If a data file contains
 * 20,000 objects with slightly different arguments to the same
 * shader we want to be able to look up the shader data
 * quickly
 */
int s_hash(char *s)
{
    unsigned int n = 0;
    while (*s)
        n = (n + *s++) % HSIZE;
    return n;
}

int s_install(shader_t *p, char *name, char *type, int slot, int dim, int block)
{
    int k = s_hash(name);
    sym_t *s = NULL;
    int ty;
    if (strcmp(type, "color") == 0)
        ty = T_COLOR;
    else if (strcmp(type, "point") == 0)
        ty = T_VECTOR;
    else if (strcmp(type, "vector") == 0)
        ty = T_VECTOR;
    else if (strcmp(type, "normal") == 0)
        ty = T_VECTOR;
    else if (strcmp(type, "float") == 0)
        ty = T_FLOAT;
    else if (strcmp(type, "string") ==0){
        if (slot < p->nstring) {
            char *null_string="";
            p->strings[slot] = (char *) POV_MALLOC(strlen(null_string)+1,
                                "string constant");
            strcpy(p->strings[slot], null_string);
        } 
        return 0;
    }
    else{
        Error("Bad type %s in s_install\n", type);
        return -1;
    }
    if (s == NULL) {
        s = (sym_t *)POV_MALLOC(sizeof(sym_t),  "shader symbol alloc");
        s->name = (char *)POV_MALLOC(strlen(name)+1, "shader symbol name alloc");
        strcpy(s->name, name);
        s->type = ty;
        s->slot = slot;
        s->block = block;
        s->next = p->htab[k];
        p->htab[k] = s;
        p->dimarr[slot] = dim;
    }
    return 0;
}


sym_t * s_lookup(shader_t *p, char *s)
{
    int k = s_hash(s);
    sym_t *q;
    for(q = p->htab[k]; q != NULL; q = q->next)
        if ((strcmp(q->name, s) == 0) && (q->block == 0))
            break;
    return q;
}

int s_getdim(shader_t *p, int offset)
{
    return p->dimarr[offset];
}

int findstring(char *s, char *d)
{
    char *p = s;
    while (*s && *s != '"')
        s++;
    if (*s != '"')
        return -1;
    s++;
    while (*s && *s != '"') {
        *d++ = *s;
        if (*s == '\\' && *(s+1) == '"')
            *d++ = *++s;
        s++;
    }
    *d = '\0';
    return s - p;
}

shader_t * s_read(FILE *fi)
{
    char lbuf[256];
    shader_t *p;
    sym_t *q;
    int n, k;
    int startaddr;
    int line =0;
    p = (shader_t *)POV_MALLOC(sizeof(shader_t), "shader allocation");
    memset(p, '\0', sizeof(shader_t));
    p->dsize = 0;

    /* TODO: allocated structures release on error!  */    
    while (fgets(lbuf, sizeof(lbuf), fi) != NULL) {
        line++;
        if (lbuf[0] == '-')
            break;
        if (lbuf[0] == '#')
            continue;
        if (lbuf[0] != ';')
            break;
        if (strncmp(lbuf+2, "version", 7) == 0) {
            int version_major=0;
            int version_minor=0;
            
            if (sscanf(lbuf+10, "%d.%d", &version_major, &version_minor) != 2){
                Error("Shader version format incorrect.\n");
            }
            else{
                if (version_major != SHADER_VERSION_MAJOR || 
                    version_minor != SHADER_VERSION_MINOR){
                    Error("Program version (%1d.%1d) and\n"
                     "shader version (%1d.%1d)differ. Recompile shader.\n",
                     SHADER_VERSION_MAJOR, SHADER_VERSION_MINOR, 
                     version_major, version_minor);    
                }
            }
        }
        else if (strncmp(lbuf+2, "requires", 8) == 0) {
            n = strtol(lbuf+10, NULL, 0);
            if (n > 0) {
                int cnt;
                p->data_templ = (DBL *)POV_MALLOC(n*sizeof(DBL),
                     "shader data vector");
                for (cnt = 0; cnt < n; cnt++){
                    p->data_templ[cnt] = 0.0;
                }
                p->dimarr = (int*)POV_MALLOC(n*sizeof(int),
                        "shader data dimension vector");
                memset(p->dimarr, 0, n*sizeof(int));
            }
            p->dsize = n;
        }
        else if (strncmp(lbuf+2, "codesize", 8) == 0) {
            n = strtol(lbuf+10, NULL, 0);
            if (n > 0) {
                p->code = (char *)POV_MALLOC(n, "shader code vector");
            }
            p->csize = n;
        }
        else if (strncmp(lbuf+2, "numstrings", 10) == 0) {
            n = strtol(lbuf+12, NULL, 0);
            if (n > 0) {
                p->strings = (char **)POV_MALLOC(n*sizeof(char *),
                        "shader string vector");
            }
            p->nstring = n;
        }
        else if (strncmp(lbuf+2, "shadername", 10) == 0) {
            char sname[128];
            sscanf(lbuf+12, "%s", sname);
            p->startaddr= 0; /* TODO: should it be read??? */
            p->name = (char *)POV_MALLOC(strlen(sname)+1,
                    "shader name string");
            strcpy(p->name, sname);
        }
        else if (strncmp(lbuf+2, "shadertype", 10) == 0) {
            int shadertype;
            sscanf(lbuf+12, "%d", &shadertype);
            p->shadertype = shadertype;
        }
        else if (strncmp(lbuf+2, "uvused", 6) == 0) {
            p->uvused++;
        }
        else if (strncmp(lbuf+2, "dynamic_string",14)==0) {
            p->dynastring++;
        }
        else if (strncmp(lbuf+2, "sym", 3) == 0) {
            char name[64], type[64];
            int slot, dim, block;
            if (p->dsize == 0) {
                Error("Shader error: missing 'requires'.\n");
                return NULL;
            }
            n = sscanf(lbuf+6, "%s %s %d %d %d",
                name, type, &slot, &dim, &block);
            if (n == 4)
                block = -1;
            else if (n != 5) {
                Error("Shader error: bad sym in line %d.\n", line);
                return NULL;
            }
            if (s_install(p, name, type, slot, dim, block) < 0)
                return NULL;
        }
        else if (strncmp(lbuf+2, "const", 5) == 0) {
            if (p->data_templ == NULL) {
                Error("Shader error: missing 'requires'.\n");
                return NULL;
            }
            if (strncmp(lbuf+8, "float", 5) == 0) {
                DBL f;
                int n;
                if (sscanf(lbuf+14, "<%lf> %d", &f, &n) == 2) {
                    if (n < p->dsize)
                        p->data_templ[n] = f;
                    else {
                        Error("Shader error: bad float const in line %d.\n", line);
                        return NULL;
                    }
                }
            } else if (strncmp(lbuf+8, "string", 6) == 0) {
                char sconst[512];
                int n, k;
                k = 15;
                k += findstring(lbuf+15, sconst);
                if (k > 15 && sscanf(lbuf+k+1, "%d", &n) == 1) {
                    if (n < p->nstring) {
                        p->strings[n] = (char *) POV_MALLOC(strlen(sconst)+1,
                                "string constant");
                        strcpy(p->strings[n], sconst);
                    } else {
                        Error("Shader error: bad string const in line %d.\n", line);
                        return NULL;
                    }
                }
            }
        }
    }
    if (p->name == NULL || p->code == NULL || p->data_templ == NULL ||
                            lbuf[0] != '-') {
        Error("Shader error: badly formed input file.\n");
        return NULL;
    }
    for(n = 0; n < p->csize; n++) {
        if (fscanf(fi, "%d", &k) != 1)
            break;
        p->code[n] = k;
    }
    if (n != p->csize) {
        Error("Shader error: short read.\n");
        return NULL;
    }

    /* set up default values for input color and opacity */
    p->data_templ[Ci_DA_OFFS  ] = 1.;  /* default output color is white */
    p->data_templ[Ci_DA_OFFS+1] = 1.;
    p->data_templ[Ci_DA_OFFS+2] = 1.;
    p->data_templ[Oi_DA_OFFS]   = 1.;  /* and opacity is opaque */
    p->data_templ[Oi_DA_OFFS+1] = 1.;
    p->data_templ[Oi_DA_OFFS+2] = 1.;
    p->data_templ[Os_DA_OFFS  ] = 1.0;
    p->data_templ[Os_DA_OFFS+1] = 1.0;
    p->data_templ[Os_DA_OFFS+2] = 1.0;
    p->data_templ[Cs_DA_OFFS  ] = 1.0;
    p->data_templ[Cs_DA_OFFS+1] = 1.0;
    p->data_templ[Cs_DA_OFFS+2] = 1.0;
    p->data_templ[As_DA_OFFS  ] = 0.1;
    p->data_templ[As_DA_OFFS+1] = 0.1;
    p->data_templ[As_DA_OFFS+2] = 0.1;
    return p;

}

SHADER * Read_Shader_File(char *fname, pov_arg_t *args, enum SHADOW_TYPE TypeOfShadow)
{
    FILE *fi;
    shader_t *p;
    shader_inst_t *q;
    pov_arg_t *a;
    sym_t *r;
    char filename_buffer[FILE_NAME_LENGTH];
    int slot = s_hash(fname);
    int i;
    for(p = shaders[slot]; p != NULL; p = p->next)
        if (strcmp(fname, p->filename) == 0)
            break;
    if (p == NULL) {
        fi = Locate_File(fname, READ_TXTFILE_STRING, ".slp", ".SLP", filename_buffer, TRUE);
        if (fi == NULL) {
            Error( "Shader reading: Can't open shader file %s.\n", fname);
        }
        p = s_read(fi);
        fclose(fi);
        if (p == NULL) {
            Error("Shader reading: shader corrupt.\n");
        }
        p->filename = (char *)POV_MALLOC(strlen(fname)+1, "shader file name");
        strcpy(p->filename, fname);
        p->next = shaders[slot];
        shaders[slot] = p;
        nloaded++;
    }
    q = (shader_inst_t *)POV_MALLOC(sizeof(shader_inst_t), "shader structure");
    memset(q, '\0', sizeof(shader_inst_t));
    q->shader = p;
    q->args = args;
    q->shadow = TypeOfShadow;
    q->Os_arg = NULL;
    q->inst_startaddr = p->startaddr;
    p->refcount++;
    if (p->nstring != 0){
        q->inst_strings = (stringtab*)POV_MALLOC(sizeof(stringtab)*p->nstring, 
            "shader string table");
        for (i=0; i < p->nstring; i++){
            q->inst_strings[i].allocated=0;
            q->inst_strings[i].string=p->strings[i];
        }
    }

    /*
     * Look up the data vector slot numbers for the instance
     * variables.  This will make a later call to the shader
     * faster.
     */
    for(a = args; a != NULL; a = a->next) {
        r = s_lookup(p, a->name);
        if (r == NULL) {
            if (strcmp(a->name, "Cs") == 0){
                a->slot = Cs_DA_OFFS;
            }
            else if (strcmp(a->name, "Os") == 0){
                a->slot = Os_DA_OFFS;
                q->Os_arg = a;
            }
            else if (strcmp(a->name, "As") == 0){
                a->slot = As_DA_OFFS;
                q->ambient_init = 1;
            }
            else{
                Error( "%s is not a valid argument to %s in shader.\n",
                    a->name, fname);
            }
        }
        else{
            if (a->type == T_COLOR && r->type == T_VECTOR)
                a->type = T_VECTOR;
            if (r->type != a->type) {
                Error("%s has the wrong type for shader %s.\n", 
                    a->name, fname);
            }
            a->slot = r->slot;
        }
    }
    return (void *)q;
}


int s_nloaded()
{
    return nloaded;
}

void reset_shader_count()
{
    nloaded = 0;
}

#endif /* ShaderPatch */

